@@ -47,6 +47,7 @@ def mkdirSafe(directory):
 
 
 def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
+                 disableMeshOptimization=False, enableMultiFile=False,
                  staticBase=False, toolSlot=None, initRotation='0 1 0 0',
                  initPos=None, linkToDef=False, jointToDef=False):
     if not os.path.exists(inFile):
@@ -61,6 +62,8 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
             sys.exit(e, '\n--init-pos argument is not valid. Your list has to be inside of quotation marks. '
                      'Example: --init-pos="1.0, 2, -0.4"')
 
+    urdf2webots.parserURDF.disableMeshOptimization = disableMeshOptimization
+    urdf2webots.writeProto.enableMultiFile = enableMultiFile
     urdf2webots.writeProto.staticBase = staticBase
     urdf2webots.writeProto.toolSlot = toolSlot
     urdf2webots.writeProto.initPos = initPos
@@ -111,6 +114,10 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
                 urdf2webots.parserURDF.robotName = robotName  # pass robotName
                 mkdirSafe(outputFile.replace('.proto', '') + '_textures')  # make a dir called 'x_textures'
 
+                if enableMultiFile:
+                    mkdirSafe(outputFile.replace('.proto', '') + '_meshes')  # make a dir called 'x_meshes'
+                    urdf2webots.writeProto.meshFilesPath = outputFile.replace('.proto', '') + '_meshes'
+
                 robot = child
                 protoFile = open(outputFile, 'w')
                 urdf2webots.writeProto.header(protoFile, inFile, robotName)
@@ -192,6 +199,11 @@ if __name__ == '__main__':
                          help='If set, the normals are exported if present in the URDF definition.')
     optParser.add_option('--box-collision', dest='boxCollision', action='store_true', default=False,
                          help='If set, the bounding objects are approximated using boxes.')
+    optParser.add_option('--disable-mesh-optimization', dest='disableMeshOptimization', action='store_true', default=False,
+                         help='If set, the duplicated vertices are not removed from the meshes (this can speed up a lot the '
+                         'conversion).')
+    optParser.add_option('--multi-file', dest='enableMultiFile', action='store_true', default=False,
+                         help='If set, the mesh files are exported as separated PROTO files')
     optParser.add_option('--static-base', dest='staticBase', action='store_true', default=False,
                          help='If set, the base link will have the option to be static (disable physics)')
     optParser.add_option('--tool-slot', dest='toolSlot', default=None,
@@ -207,5 +219,5 @@ if __name__ == '__main__':
     optParser.add_option('--joint-to-def', dest='jointToDef', action='store_true', default=False,
                          help='If set, urdf joint names are also used as DEF names as well as joint names.')
     options, args = optParser.parse_args()
-    convert2urdf(options.inFile, options.outFile, options.normal, options.boxCollision, options.staticBase,
-                 options.toolSlot, options.initRotation, options.initPos, options.linkToDef, options.jointToDef)
+    convert2urdf(options.inFile, options.outFile, options.normal, options.boxCollision, options.disableMeshOptimization,
+                 options.enableMultiFile, options.staticBase, options.toolSlot, options.initRotation, options.initPos, options.linkToDef, options.jointToDef)
